import copy
from typing import Union

import numpy as np
import pandas as pd
import pywt
from scipy import signal
from scipy.stats import zscore

from . import functional as F


__all__ = [
    "FrequencyResample",
    "Padding",
    "EdgeCut",
    "Normalization",
    "ButterworthFilter",
    "IIRNotchFilter",
    "BaselineWanderRemoval",
    "WaveletTransform",
    "LeadCrop",
    "RandomLeadCrop",
    "TimeCrop",
    "RandomTimeCrop",
    "SumAug",
    "RandomSumAug",
    "ReflectAug",
    "ConvexAug",
    "RandomConvexAug",
]


class FrequencyResample:
    """
    Apply frequency resample
    :param ecg_frequency: sampling frequency of a signal
    :param requested_frequency: sampling frequency of a preprocessed signal

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        ecg_frequency: int, 
        requested_frequency: int = 500,
    ):
        if isinstance(ecg_frequency, (int, float)):
            self.ecg_frequency = ecg_frequency
        else:
            raise ValueError('ecg_frequency must be scalar')
        if isinstance(requested_frequency, (int, float)):
            self.requested_frequency = requested_frequency
        else:
            raise ValueError('requested_frequency must be scalar')
        self.func = F.ecg_to_one_frequency

    def __call__(self, x):

        x = self.func(x, int(self.ecg_frequency), int(self.requested_frequency))
        
        return x
    

class Padding:
    """
    Apply padding. If ECG is longer than the observed_ecg_length the record is cut.
    :param observed_ecg_length: length of padded signal in seconds
    :param frequency: sampling frequency of a signal
    :param pad_mode: padding mode

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        observed_ecg_length: float = 10, 
        frequency: int = 500, 
        pad_mode: str = "constant",
    ):
        self.observed_ecg_length = observed_ecg_length
        self.frequency = frequency
        self.pad_mode = pad_mode

    def __call__(self, x):

        if self.observed_ecg_length*self.frequency - x.shape[1] > 0:
            x = np.pad(x, ((0, 0), (0, self.observed_ecg_length*self.frequency - x.shape[1])), mode=self.pad_mode)
        else:
            x = x[:, :self.observed_ecg_length*self.frequency]
            
        return x


class EdgeCut:
    """
    Cut signal edges 
    :param cut_range: cutting parameters
    :param frequency: sampling frequency of a signal

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        cut_range: list = [0, 0], 
        frequency: int = 500,
    ):
        self.cut_range = cut_range
        self.frequency = frequency
        self.func = F.cut_ecg
        
    def __call__(self, x):

        x = self.func(x, self.cut_range, self.frequency)
        
        return x
        
    
class Normalization:
    """
    Apply normalization
    :param norm_type: type of normalization ('z_norm' or 'cycle') 
    :param leads: leads to be normalized

    :return: preprocessed data
    """

    def __init__(
        self, 
        norm_type: str = "z_norm", 
        leads: list = None,
    ):
        if leads is None:
            self.leads = list(range(12))
        elif isinstance(leads, list):
            self.leads = leads
        else:
            raise ValueError('leads must be list type')
        self.norm_type = norm_type
        if norm_type == "cycle":
            self.func = F.cycle_normalization
        elif norm_type == "min_max":
            self.func = F.minmax_normalization
        elif norm_type == "z_norm":
            self.func = F.z_normalization
        else:
            raise ValueError('norm_type must be one of [cycle, min_max, z_norm]')

    def __call__(self, x):
        
        if self.norm_type is not None:
            return self.func(x[self.leads, :])
        else:
            return x[self.leads, :]

    
class ButterworthFilter:
    """
    Apply Butterworth filter augmentation
    :param filter_type: type of Butterworth filter ('bandpass', 'lowpass' or 'highpass') 
    :param leads: leads to be filtered
    :param n: filter order
    :param Wn: cutoff frequency(ies)
    :param fs: filtered signal frequency

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        filter_type: str = "bandpass", 
        leads: list = None, 
        n: int = 10, 
        Wn: Union[float,list] = [3, 30], 
        fs: int = 500,
    ):
        if leads is None:
            self.leads = list(range(12))
        elif isinstance(leads, list):
            self.leads = leads
        else:
            raise ValueError('leads must be list type')
        self.filter_type = filter_type
        if filter_type == "bandpass":
            self.func = F.butterworth_bandpass_filter
        elif filter_type == "lowpass":
            self.func = F.butterworth_lowpass_filter
        elif filter_type == "highpass":
            self.func = F.butterworth_highpass_filter
        else:
            raise ValueError("Filter type must be one of [bandpass, lowpass, highpass]")
        self.n = n
        if filter_type == "bandpass" and not isinstance(Wn, list):
            raise ValueError('Wn must be list type in case of bandpass filter')
        elif (filter_type == "highpass" or filter_type == "lowpass") and not isinstance(Wn, (int, float)):
            raise ValueError(f'Wn must be a scalar in case of {filter_type} filter')
        self.Wn = Wn
        self.fs = fs

    def __call__(self, x):
        
        return np.apply_along_axis(self.func, axis=1, arr=x[self.leads, :], n=self.n, Wn=self.Wn, fs=self.fs)

    
class IIRNotchFilter:
    """
    Apply IIR notch filter augmentation
    :param leads: leads to be filtered
    :param w0: frequency to remove from a signal
    :param Q: quality factor
    :param fs: sampling frequency of a signal

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        leads: list = None, 
        w0: float = 50, 
        Q: float = 30, 
        fs: int = 500,
    ):
        if leads is None:
            self.leads = list(range(12))
        elif isinstance(leads, list):
            self.leads = leads
        else:
            raise ValueError('leads must be list type')
        self.w0 = w0
        self.Q = Q
        self.fs = fs
        self.func = F.IIR_notch_filter

    def __call__(self, x):
        
        return np.apply_along_axis(self.func, axis=1, arr=x[self.leads, :], w0=self.w0, Q=self.Q, fs=self.fs)


class BaselineWanderRemoval:
    """
    Remove baseline wander using wavelets (see article https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.308.6789&rep=rep1&type=pdf)
    :param leads: leads to be preprocessed
    :param wavelet: wavelet name

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        leads: list = None, 
        wavelet: str = 'db4',
    ):
        if leads is None:
            self.leads = list(range(12))
        elif isinstance(leads, list):
            self.leads = leads
        else:
            raise ValueError('leads must be list type')
        self.wavelet = wavelet
        self.func = F.DWT_BW

    def __call__(self, x):
        
        return np.apply_along_axis(self.func, axis=1, arr=x[self.leads, :], wavelet=self.wavelet)

    
class WaveletTransform:
    """
    Apply wavelet transform augmentation
    :param wt_type: type of wavelet transform ('DWT' with soft thresholding or 'SWT') 
    :param leads: leads to be transformed
    :param wavelet: wavelet name
    :param level: decomposition level
    :param threshold: thresholding value for all coefficients except the first one (only for DWT)
    :param low: thresholding value for the first coefficient (only for DWT)

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        wt_type: str = "DWT", 
        leads: list = None, 
        wavelet: str = 'db4', 
        level: int = 3, 
        threshold: float = 2, 
        low: float = 1e6,
    ):
        if leads is None:
            self.leads = list(range(12))
        elif isinstance(leads, list):
            self.leads = leads
        else:
            raise ValueError('leads must be list type')
        self.wt_type = wt_type
        self.wavelet = wavelet
        self.level = level
        if wt_type == "DWT":
            self.threshold = threshold
            self.low = low
            self.func = F.DWT_filter
        elif wt_type == "SWT":
            self.threshold = None
            self.low = None
            self.func = F.SWT_filter
        else:
            raise ValueError('wt_type must be one of [DWT, SWT]')
            
    def __call__(self, x):
        
        if self.wt_type == "DWT":
            return np.apply_along_axis(self.func, axis=1, arr=x[self.leads, :], wavelet=self.wavelet, 
                                       level=self.level, threshold=self.threshold, low=self.low)
        elif self.wt_type == "SWT":
            return np.apply_along_axis(self.func, axis=1, arr=x[self.leads, :], wavelet=self.wavelet, 
                                       level=self.level)


class LeadCrop:
    """
    Apply lead crop augmentation
    :param leads: leads to be cropped

    :return: preprocessed data
    """
    
    def __init__(
        self,
        leads: list = [0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    ):
        self.leads = leads
        self.func = F.lead_crop
        
    def __call__(self, x):
        
        return self.func(x, leads=self.leads)
    
    
class RandomLeadCrop(LeadCrop):
    """
    Apply random lead crop augmentation
    :param n: number of leads to be cropped (chosen randomly)

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        n: int = 11,
    ):
        ls = np.arange(12, dtype='int')
        leads_to_remove = np.random.choice(ls, size=n, replace=False)
        super().__init__(leads_to_remove)
        self.n = n

    def __call__(self, x):
        
        return self.func(x, leads=self.leads)

    
class TimeCrop:
    """
    Apply time crop augmentation
    :param time: length of time segment to be cropped (the same units as signal)
    :param leads: leads to be cropped

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        time: int = 100, 
        leads: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    ):
        self.time = time
        self.leads = leads
        self.func = F.time_crop
        
    def __call__(self, x):
        
        return self.func(x, time=self.time, leads=self.leads)

    
class RandomTimeCrop(TimeCrop):
    """
    Apply random time crop augmentation
    :param time: length of time segment to be cropped (the same units as signal)
    :param n: number of leads to be cropped (chosen randomly)

    :return: preprocessed data
    """
    
    def __init__(
        self, 
        time: int = 100, 
        n: int = 12,
    ):
        ls = np.arange(12, dtype='int')
        leads_to_modify = np.random.choice(ls, size=n, replace=False)
        super().__init__(time, leads_to_modify)
        self.time = time
        self.n = n
        
    def __call__(self, x):

        return self.func(x, time=self.time, leads=self.leads)
    
    
class SumAug:
    """
    Apply sum augmentation to selected leads
    :param leads: leads to be replaced by sum of all leads

    :return: preprocessed data
    """
    
    def __init__(
        self,
        leads: list = [0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    ):
        self.leads = leads
        self.func = F.sum_augmentation
        
    def __call__(self, x):
        
        return self.func(x, leads=self.leads)
    
    
class RandomSumAug(SumAug):
    """
    Apply random sum augmentation
    :param n: number of leads to be replaced by sum of all leads (chosen randomly)

    :return: preprocessed data
    """
    
    def __init__(
        self,
        n: int = 11,
    ):
        ls = np.arange(12, dtype='int')
        leads_to_remove = np.random.choice(ls, size=n, replace=False)
        super().__init__(leads_to_remove)
        self.n = n
        
    def __call__(self, x):
        
        return self.func(x, leads=self.leads)
    
    
class ReflectAug:
    """
    Apply reflection augmentation

    :return: preprocessed data
    """
    
    def __init__(
        self,
    ):
        self.func = F.reflect_augmentation
        
    def __call__(self, x):
        
        return self.func(x)
    
    
class ConvexAug:
    """
    Apply convex augmentation
    :param leads: leads to be replaced by convex combination of some leads (chosen randomly)

    :return: preprocessed data
    """
    
    def __init__(
        self,
        leads: list = [0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    ):
        self.leads = leads
        self.func = F.convex_augmentation
        
    def __call__(self, x):
        
        return self.func(x, leads=self.leads)
    
    
class RandomConvexAug(ConvexAug):
    """
    Apply random convex augmentation
    :param n: number of leads (chosen randomly) to be replaced by convex combination of some leads (chosen randomly)

    :return: preprocessed data
    """
    
    def __init__(
        self,
        n: int = 11,
    ):
        ls = np.arange(12, dtype='int')
        leads_to_remove = np.random.choice(ls, size=n, replace=False)
        super().__init__(leads_to_remove)
        self.n = n
        
    def __call__(self, x):
        
        return self.func(x, leads=self.leads)


import random

import numpy as np


__all__ = [
    "Compose",
    "SomeOf",
    "OneOf",
]


class Compose:
    """
    Apply successively a list of preprocessing techniques (normalization, augmentation etc.)
    :param transforms: a list of preprocessing techniques
    :param p: probability of preprocessing pipeline

    :return: preprocessed data
    """

    def __init__(
        self,
        transforms: list,
        p: float,
    ):
        self.transforms = transforms
        if p > 1.0 or p < 0.0:
            raise ValueError("Probability should be between 0.0 and 1.0")
        else:
            self.p = p

    def __call__(self, x):
        
        idx = np.random.RandomState(random.randint(0, (1 << 32) - 1)).choice(2, p=[self.p, 1.0-self.p])
        if idx == 0:
            for t in self.transforms:
                x = t(x)

        return x


class SomeOf:
    """
    Apply some of preprocessing techniques from a given list (normalization, augmentation etc.)
    :param transforms: a list of preprocessing techniques
    :param n: a number of preprocessing techniques to apply
    :param transform_prob: a list of probabilities

    :return: preprocessed data
    """

    def __init__(
        self,
        transforms: list,
        n: int,
        transform_prob: list = None,
    ):
        self.transforms = transforms
        self.n = n
        if transform_prob is None:
            self.transform_prob = [1 / len(self.transforms)]*len(self.transforms)
        else:
            if sum(transform_prob) > 1.0:
                raise ValueError("Sum of probabilities should be equal to 1.0")
            else:
                self.transform_prob = transform_prob

    def __call__(self, x):
        
        idx = np.random.RandomState(random.randint(0, (1 << 32) - 1)).choice(len(self.transforms), size=self.n, p=self.transform_prob, replace=False)
        for i in idx:
            t = self.transforms[i]
            data = t(x)
            
        return data
    

class OneOf:
    """
    Apply one of preprocessing techniques from a given list (normalization, augmentation etc.)
    :param transforms: a list of preprocessing techniques
    :param transform_prob: a list of probabilities

    :return: preprocessed data
    """
    
    def __init__(
        self,
        transforms: list,
        transform_prob: list = None,
    ):
        self.transforms = transforms
        if transform_prob is None:
            self.transform_prob = [1 / len(self.transforms)]*len(self.transforms)
        else:
            if sum(transform_prob) > 1.0:
                raise ValueError("Sum of probabilities should be equal to 1.0")
            else:
                self.transform_prob = transform_prob

    def __call__(self, x):
        
        idx = np.random.RandomState(random.randint(0, (1 << 32) - 1)).choice(len(self.transforms), p=self.transform_prob)
        t = self.transforms[idx]
        x = t(x)
        
        return x


from enum import IntEnum
from pathlib import Path
from typing import Callable, Optional

import ecg_plot
import numpy as np
import pandas as pd
import torch
import wfdb
from ecglib import preprocessing as P
from torch.utils.data import Dataset

__all__ = [
    "EcgDataset",
    "TisDataset",
    "PTBXLDataset",
]


class EcgDataset(Dataset):
    """
    EcgDataset
    :param ecg_data: dataframe with ecg info
    :param target: a list of targets
    :param frequency: frequency for signal resampling
    :param leads: a list of leads
    :param ecg_length: length of ECG signal after padding / cropping
    :param cut_range: cutting parameters
    :param pad_mode: padding mode 
    :param classes: number of classes
    :param use_meta: whether to use metadata or not
    :param augmentation: a bunch of augmentations and other preprocessing techniques
    """

    def __init__(
        self,
        ecg_data: pd.DataFrame,
        target: list,
        frequency: int = 500,
        leads: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        ecg_length: float = 10,
        cut_range: list = [0, 0],
        pad_mode: str = "constant",
        norm_type: str = "z_norm",
        classes: int = 2,
        use_meta: bool = False,
        augmentation: Callable = None,
    ):
        super().__init__()
        if "fpath" not in ecg_data.columns:
            raise ValueError("column 'fpath' not in ecg_data")
        self.ecg_data = ecg_data
        self.target = target
        self.frequency = frequency
        self.leads = leads
        self.ecg_length = ecg_length
        self.cut_range = cut_range
        self.pad_mode = pad_mode
        self.norm_type = norm_type
        self.classes = classes
        if use_meta and "ecg_parameters" not in ecg_data.columns:
            raise ValueError("metadata column 'ecg_parameters' not in ecg_data")
        self.meta = use_meta
        self.augmentation = augmentation

    def __len__(self):
        return self.ecg_data.shape[0]

    def get_fpath(self, index: int) -> str:
        """
        Returns path to file with ECG leads
        :param index: Index of ECG in dataset
        
        :return: Path to ECG file
        """

        return self.ecg_data.iloc[index]["fpath"]

    def get_name(self, index: int) -> str:
        """
        Returns name of ECG file
        :param index: Index of ECG in dataset
        
        :return: ECG file name
        """

        return str(Path(self.get_fpath(index)).stem)


class TisDataset(EcgDataset):
    """
    TisDataset
    :param ecg_data: dataframe with ecg info
    :param target: a list of targets
    :param frequency: frequency for signal resampling
    :param leads: a list of leads
    :param ecg_length: length of ECG signal after padding / cropping
    :param cut_range: cutting parameters
    :param pad_mode: padding mode 
    :param classes: number of classes
    :param use_meta: whether to use metadata or not
    :param augmentation: a bunch of augmentations and other preprocessing techniques
    """

    def __init__(
        self,
        ecg_data: pd.DataFrame,
        target: list,
        frequency: int = 500,
        leads: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        ecg_length: float = 10,
        cut_range: list = [0, 0],
        pad_mode: str = "constant",
        norm_type: str = "z_norm",
        classes: int = 2,
        use_meta: bool = False,
        augmentation: Callable = None,
    ):
        super().__init__(
            ecg_data,
            target,
            frequency,
            leads,
            ecg_length,
            cut_range,
            pad_mode,
            norm_type,
            classes,
            use_meta,
            augmentation,
        )

    def __getitem__(self, index):

        if "frequency" in self.ecg_data.columns:
            ecg_frequency = float(self.ecg_data.iloc[index]["frequency"])
        else:
            ecg_frequency = self.frequency

        # data standartization (scaling, resampling, cuts off, normalization and padding/truncation)
        ecg_record = np.load(self.ecg_data.iloc[index]["fpath"])["arr_0"].astype(
            "float64"
        )
        ecg_record = P.Compose(
            transforms=[
                P.FrequencyResample(
                    ecg_frequency=ecg_frequency, requested_frequency=self.frequency
                ),
                P.EdgeCut(cut_range=self.cut_range, frequency=self.frequency),
                P.Normalization(norm_type=self.norm_type),
                P.Padding(
                    observed_ecg_length=self.ecg_length, frequency=self.frequency
                ),
            ],
            p=1.0,
        )(ecg_record)
        assert not np.any(
            np.isnan(ecg_record)
        ), f"ecg_record = {ecg_record}, index = {index}"

        # data preprocessing if specified (augmentation, filtering)
        if self.augmentation is not None:
            ecg_record = self.augmentation(ecg_record)

        target = self.target[index]

        result = [
            torch.tensor(ecg_record[self.leads, :], dtype=torch.float),
            torch.tensor(target, dtype=torch.float),
        ]

        if self.meta:
            result.append(
                torch.tensor(
                    self.ecg_data.iloc[index]["ecg_parameters"], dtype=torch.float
                )
            )

        return (index, result)

    def save_as_png(
        self, index: int, dest_path: str, postfix: Optional[str] = None
    ) -> None:
        """
        Saves the image of ecg record

        :param index: Index of ECG
        :param dest_path: Directory to save the image
        :param postfix: Subdirectory where the image will be saved, defaults to None
        """

        ecg = (np.load(self.get_fpath(index))["arr_0"].astype("float64"),)
        ecg = np.squeeze(ecg)

        if "frequency" in self.ecg_data.columns:
            frequency = self.ecg_data.iloc[index]["frequency"]
        else:
            frequency = self.frequency
        ecg = ecg / np.max(
            ecg
        )  # added to fit the record to the visible part of the plot
        ecg_plot.plot(ecg, sample_rate=frequency)
        ecg_fname = self.get_name(index)

        if postfix:
            dest_path = str(Path(dest_path).joinpath(postfix))

        dest_path = (
            "{}/".format(dest_path) if not dest_path.endswith("/") else dest_path
        )

        if not Path(dest_path).exists():
            Path(dest_path).mkdir(parents=True, exist_ok=True)

        ecg_plot.save_as_png(file_name=ecg_fname, path=dest_path)

    @staticmethod
    def for_train_from_config(
        data: pd.DataFrame, target: list, augmentation: Callable, config: dict, classes_num: int
    ) -> EcgDataset:
        """
        A wrapper with just four parameters to create `TisDataset` for training and validation
        :param data: dataframe with ecg info
        :param target: a list of targets
        :param augmentation: a bunch of augmentations and other preprocessing techniques
        :param config: config dictionary
        :param classes_num: number of classes
        
        :return: TisDataset
        """

        return TisDataset(
            data,
            target,
            frequency=config.ecg_record_params.resampled_frequency,
            leads=config.ecg_record_params.leads,
            ecg_length=config.ecg_record_params.observed_ecg_length,
            norm_type=config.ecg_record_params.normalization,
            classes=classes_num,
            use_meta=config.ecg_metadata.use_metadata,
            cut_range=config.ecg_record_params.ecg_cut_range,
            augmentation=augmentation,
        )


class PTBXLDataset(EcgDataset):
    """
    PTBXLDataset
    :param ecg_data: dataframe with ecg info
    :param target: a list of targets
    :param frequency: frequency for signal resampling
    :param leads: a list of leads
    :param ecg_length: length of ECG signal after padding / cropping
    :param cut_range: cutting parameters
    :param pad_mode: padding mode 
    :param classes: number of classes
    :param use_meta: whether to use metadata or not
    :param augmentation: a bunch of augmentations and other preprocessing techniques
    """

    def __init__(
        self,
        ecg_data: pd.DataFrame,
        target: list,
        frequency: int = 500,
        leads: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        ecg_length: float = 10,
        cut_range: list = [0, 0],
        pad_mode: str = "constant",
        norm_type: str = "z_norm",
        classes: int = 2,
        use_meta: bool = False,
        augmentation: Callable = None,
    ):
        super().__init__(
            ecg_data,
            target,
            500,
            leads,
            10,
            cut_range,
            pad_mode,
            norm_type,
            classes,
            False,
            augmentation,
        )

    def __getitem__(self, index):
        if "frequency" in self.ecg_data.columns:
            ecg_frequency = float(self.ecg_data.iloc[index]["frequency"])
        else:
            ecg_frequency = self.frequency

        ecg_record, _ = wfdb.rdsamp(
            self.ecg_data.iloc[index]["fpath"], channels=self.leads
        )
        ecg_record = ecg_record.T
        ecg_record = ecg_record.astype("float64")

        # data standartization (scaling, resampling, cuts off, normalization and padding/truncation)
        ecg_record = P.Compose(
            transforms=[
                P.FrequencyResample(
                    ecg_frequency=ecg_frequency, requested_frequency=self.frequency
                ),
                P.EdgeCut(cut_range=self.cut_range, frequency=self.frequency),
                P.Normalization(norm_type=self.norm_type),
                P.Padding(
                    observed_ecg_length=self.ecg_length, frequency=self.frequency
                ),
            ],
            p=1.0,
        )(ecg_record)

        # data preprocessing if specified (augmentation, filtering)
        if self.augmentation is not None:
            ecg_record = self.augmentation(ecg_record)

        target = self.target[index]

        result = [
            torch.tensor(ecg_record[self.leads, :], dtype=torch.float),
            torch.tensor(target, dtype=torch.float),
        ]

        return (index, result)

    def save_as_png(
        self, index: int, dest_path: str, postfix: Optional[str] = None
    ) -> None:
        """
        Saves the image of ecg record
        :param index: Index of ECG
        :param dest_path: Directory to save the image
        :param postfix: Subdirectory where the image will be saved, defaults to None
        """

        if "frequency" in self.ecg_data.columns:
            frequency = self.ecg_data.iloc[index]["frequency"]
        else:
            frequency = self.frequency

        ecg, _ = wfdb.rdsamp(self.ecg_data.iloc[index]["fpath"], channels=self.leads)
        ecg = ecg.T
        ecg = ecg.astype("float64")

        ecg = ecg / np.max(
            ecg
        )  # added to fit the record to the visible part of the plot
        ecg_plot.plot(ecg, sample_rate=frequency)
        ecg_fname = self.get_name(index)

        if postfix:
            dest_path = str(Path(dest_path).joinpath(postfix))

        dest_path = (
            "{}/".format(dest_path) if not dest_path.endswith("/") else dest_path
        )

        if not Path(dest_path).exists():
            Path(dest_path).mkdir(parents=True, exist_ok=True)

        ecg_plot.save_as_png(file_name=ecg_fname, path=dest_path)

    @staticmethod
    def for_train_from_config(
        data: pd.DataFrame, target: list, augmentation: Callable, config: dict, classes_num: int
    ) -> EcgDataset:
        """
        A wrapper with just four parameters to create `PTBXLDataset` for training and validation
        :param data: dataframe with ecg info
        :param target: a list of targets
        :param augmentation: a bunch of augmentations and other preprocessing techniques
        :param config: config dictionary
        :param classes_num: number of classes
        
        :return: PTBXLDataset
        """

        return PTBXLDataset(
            data,
            target,
            frequency=config.ecg_record_params.resampled_frequency,
            leads=config.ecg_record_params.leads,
            ecg_length=config.ecg_record_params.observed_ecg_length,
            norm_type=config.ecg_record_params.normalization,
            classes=classes_num,
            use_meta=config.ecg_metadata.use_metadata,
            cut_range=config.ecg_record_params.edge_cut.ecg_cut_range,
            augmentation=augmentation,
        )




import os
from collections import OrderedDict
from dataclasses import asdict
from operator import mod

import hydra
import torch
from omegaconf import DictConfig
import yaml

from .architectures.densenet1d import densenet121_1d, densenet201_1d
from .config.model_configs import DenseNetConfig

resource_package = __name__

pathologies = ["AFIB", "STACH", "SBRAD", "RBBB", "LBBB", "PVC", "1AVB"]

arch_map = {
    "densenet1d121": densenet121_1d,
    "densenet1d201": densenet201_1d,
}


def get_config(model_name: str, config: dict = None):
    if config:
        model_name = config.model_name
        return hydra.utils.instantiate(config.config)

    if "densenet" in model_name:
        return DenseNetConfig()
    else:
        raise Exception("Unknown model type")


def get_model(
    model_name: str,
    leads_num: int,
    model_cfg: DictConfig,
    num_classes: int,
) -> torch.nn.Module:
    if model_cfg:
        model_name = model_cfg.model_name

    assert (
        model_name in arch_map
    ), "Model name must be one of ['densenet1d121', 'densenet1d201']"

    model_config = get_config(model_name=model_name, config=model_cfg)
    model_config.input_channels = leads_num
    model_config.num_classes = num_classes

    # Note: Overloads config params according to input args values
    if isinstance(model_config, DenseNetConfig):
        return arch_map[model_name](**asdict(model_config))


def create_model(
    model_name: str,
    pathology: str,
    model_cfg: dict = None,
    pretrained: bool = False,
    leads_num: int = 12,
):
    model = get_model(
        model_name=model_name,
        leads_num=leads_num,
        model_cfg=model_cfg,
        num_classes=1,
        # meta=meta,
    )

    if pretrained:
        if pathology not in pathologies:
            raise KeyError(
                "pathology must be one of ['AFIB', 'STACH', 'SBRAD', 'RBBB', 'LBBB', 'PVC', '1AVB']"
            )

        dirname = os.path.dirname(__file__)
        weights_path = os.path.join(dirname, 'weights/model_weights_paths.yaml')
        with open(weights_path, 'r') as file:
            weights_config = yaml.safe_load(file)
        if model_name not in weights_config[f"{leads_num}_leads"][pathology]:
            raise KeyError(
                "the weights are currently available for the following architectures ['densenet1d121']"
            )
        weights_path = weights_config[f"{leads_num}_leads"][pathology][model_name]

        model_info = torch.hub.load_state_dict_from_url(weights_path, progress=True, check_hash=True, file_name=f"{leads_num}_leads_{model_name}_{pathology}.pt")

        print('{} model trained on {}-lead {} second ECG records with {} frequency to detect {}. {} normalization was applied to all the records during preprocessing.'.format(
            model_name,
            leads_num,
            model_info['config_file']['ecg_record_params']['observed_ecg_length'],
            model_info['config_file']['ecg_record_params']['resampled_frequency'],
            pathology,
            model_info['config_file']['ecg_record_params']['normalization'],
        ))


        model_state_dict = model_info['model']
        model.load_state_dict(model_state_dict)
    return model



import copy

import numpy as np
import pywt
from scipy import signal
from scipy.stats import zscore


__all__ = [
    "butterworth_bandpass_filter",
    "butterworth_highpass_filter",
    "butterworth_lowpass_filter",
    "IIR_notch_filter",
    "elliptic_bandpass_filter",
    "minmax_normalization",
    "z_normalization",
    "cycle_normalization",
    "DWT_filter",
    "SWT_filter",
    "lead_crop",
    "time_crop",
    "sum_augmentation",
    "convex_augmentation",
    "reflect_augmentation",
    "ecg_to_one_frequency",
    "DWT_BW",
    "cut_ecg",
]


def butterworth_bandpass_filter(
    s: np.ndarray, 
    n: int = 10, 
    Wn: list = [3, 30], 
    fs: float = 500,
) -> np.ndarray:
    """
    Butterworth bandpass filter augmentation
    :param s: one lead signal
    :param n: filter order
    :param Wn: cutoff frequencies
    :param fs: filtered signal frequency

    :return: preprocessed data
    """
    
    sos = signal.butter(N=n, btype='bandpass', Wn=Wn, fs=fs, output='sos')
    filtered_signal = signal.sosfiltfilt(sos, s)
    return filtered_signal


def butterworth_highpass_filter(
    s: np.ndarray, 
    n: int = 7, 
    Wn: float = 0.5, 
    fs: int = 500,
) -> np.ndarray:
    """
    Butterworth highpass filter augmentation
    :param s: one lead signal
    :param n: filter order
    :param Wn: cutoff frequency
    :param fs: filtered signal frequency

    :return: preprocessed data
    """
    
    sos = signal.butter(N=n, btype='highpass', Wn=Wn, fs=fs, output='sos')
    filtered_signal = signal.sosfiltfilt(sos, s)
    return filtered_signal


def butterworth_lowpass_filter(
    s: np.ndarray, 
    n: int = 6, 
    Wn: float = 20, 
    fs: int = 500,
) -> np.ndarray:
    """
    Butterworth lowpass filter augmentation
    :param s: one lead signal
    :param n: filter order
    :param Wn: cutoff frequency
    :param fs: filtered signal frequency

    :return: preprocessed data
    """
    
    sos = signal.butter(N=n, btype='lowpass', Wn=Wn, fs=fs, output='sos')
    filtered_signal = signal.sosfiltfilt(sos, s)
    return filtered_signal


def IIR_notch_filter(
    s: np.ndarray, 
    w0: float = 50, 
    Q: float = 30, 
    fs: int = 500,
) -> np.ndarray:
    """
    IIR notch filter augmentation
    :param s: one lead signal
    :param w0: frequency to remove from a signal
    :param Q: quality factor
    :param fs: sampling frequency of a signal

    :return: preprocessed data
    """
    
    b, a = signal.iirnotch(w0=w0, Q=Q, fs=fs)
    filtered_signal = signal.filtfilt(b, a, s)
    return filtered_signal


def elliptic_bandpass_filter(
    s: np.ndarray, 
    n: int = 10, 
    rp: float = 4, 
    rs: float = 5, 
    Wn: list = [0.5, 50], 
    fs: int = 500,
) -> np.ndarray:
    """
    Elliptic bandpass filter
    :param s: one lead signal
    :param n: filter order
    :param rp: maximum ripple allowed below unity gain in the passband
    :param rs: minimum attenuation required in the stop band
    :param Wn: cutoff frequencies
    :param fs: filtered signal frequency
    
    :return: preprocessed data
    """
    
    sos = signal.ellip(N=n, btype='bandpass', rp=rp, rs=rs, Wn=Wn, fs=fs, output='sos')
    filtered_signal = signal.sosfiltfilt(sos, s)
    return filtered_signal


def minmax_normalization(
    s: np.ndarray,
) -> np.ndarray:
    """
    minmax_normalization
    :param s: signal

    :return: preprocessed data
    """
    
    return (s-np.min(s))/(np.max(s)-np.min(s))


def z_normalization(
    s: np.ndarray,
) -> np.ndarray:
    """
    Z-normalization
    :param s: signal

    :return: preprocessed data
    """
    
    return zscore(s, axis=1, nan_policy='raise')


def cycle_normalization(
    s: np.ndarray,
) -> np.ndarray:
    """
    Cycle normalization
    :param s: one lead signal

    :return: preprocessed data
    """

    smin = np.min(s)
    smax = np.max(s)
    n = len(s) - 1
    i = np.arange(len(s))
    return ((n - i) / n) * ((s - smin) / (smax - smin)) + (i / n) * ((s - smin) / (smax - smin))


def DWT_filter(
    s: np.ndarray, 
    wavelet: str = 'db4', 
    level: int = 3, 
    threshold: float = 2, 
    low: float = 1e6,
) -> np.ndarray:
    """
    Discrete wavelet transform augmentation
    :param s: one lead signal
    :param wavelet: wavelet name
    :param level: decomposition level
    :param threshold: thresholding value for all coefficients except the first one
    :param low: thresholding value for the first coefficient

    :return: preprocessed data
    """
    
    w = pywt.Wavelet(wavelet)
    maxlev = pywt.dwt_max_level(len(s), w.dec_len)

    assert maxlev >= level

    coeffs = pywt.wavedec(s, w, level=level)
    for i in range(1, len(coeffs)):
        coeffs[i] = pywt.threshold(coeffs[i], threshold * np.sqrt(np.log2(len(coeffs[i]))), mode='soft')
    coeffs[0] = pywt.threshold(coeffs[0], low, mode='less')
    return pywt.waverec(coeffs, wavelet, mode='periodic')


def SWT_filter(
    s: np.ndarray, 
    wavelet: str = 'db4', 
    level: int = 6,
) -> np.ndarray:
    """
    Stationary wavelet transform augmentation
    :param s: one lead signal
    :param wavelet: wavelet name
    :param level: decomposition level

    :return: preprocessed data
    """
    
    if len(s) % 2 == 0:
        width = (2**int(np.ceil(np.log2(len(s)))) - len(s)) // 2
        s_padded = np.pad(s, pad_width=width, mode='symmetric')
    else:
        width1 = (2**int(np.ceil(np.log2(len(s)))) - len(s)) // 2
        width2 = (2**int(np.ceil(np.log2(len(s)))) - len(s)) // 2 + 1
        s_padded = np.pad(s, pad_width=(width1, width2), mode='symmetric')
    
    w = pywt.Wavelet(wavelet)
    maxlev = pywt.swt_max_level(len(s_padded))

    assert maxlev >= level

    coeffs = pywt.swt(s_padded, w, level=level, trim_approx=True, norm=True)
    return pywt.iswt(coeffs, wavelet, norm=True)


def lead_crop(
    record: np.ndarray, 
    leads: list,
) -> np.ndarray:
    """
    Lead crop augmentation
    :param record: signal
    :param leads: leads to be cropped

    :return: preprocessed data
    """
    
    record[leads, :] = np.zeros(record.shape[1])
    return record


def time_crop(
    record: np.ndarray, 
    time: int, 
    leads: list,
) -> np.ndarray:
    """
    Time crop augmentation
    :param record: signal
    :param time: length of time segment to be cropped (the same units as signal)
    :param leads: leads to be cropped

    :return: preprocessed data
    """
    
    assert time <= record.shape[1]
    for lead in leads:
        ls = np.arange(0, len(record[lead])-time, dtype="int")
        start = np.random.choice(ls)
        record[lead, start:start+time] = 0
    return record


def sum_augmentation(
    record: np.ndarray,
    leads: list,
) -> np.ndarray:
    """
    Signal summation augmentation
    
    :param record: signal
    :param leads: leads to be replaced by sum of all leads

    :return: preprocessed data
    """
    
    record[leads, :] = np.sum(record, axis=0)
    return record


def reflect_augmentation(
    record: np.ndarray,
) -> np.ndarray:
    """
    Reflection augmentation
    
    :param record: signal

    :return: preprocessed data
    :rtype: numpy 2d array
    """
    
    return -record


def convex_augmentation(
    record: np.ndarray,
    leads: list,
) -> np.ndarray:
    """
    Convex augmentation
    
    :param record: signal
    :param leads: leads to be replaced by convex combination of some leads (chosen randomly)

    :return: preprocessed data
    """
    
    result = copy.deepcopy(record)
    ls = np.arange(12, dtype='int')
    for ltr in leads:
        ln = np.random.randint(12, size=1)[0]
        leads_to_sum = np.random.choice(ls, size=ln, replace=False)
        convex_coeffs = np.random.dirichlet(np.ones(ln), size=1)[0]
        if ln != 0:
            result[ltr, :] = np.dot(convex_coeffs, record[leads_to_sum, :])
    return result


def ecg_to_one_frequency(
    ecg_record: np.ndarray, 
    ecg_frequency: int, 
    requested_frequency: int,
) -> np.ndarray:
    """
    Frequency resample
    :param record: signal
    :param ecg_frequency: sampling frequency of a signal
    :param requested_frequency: sampling frequency of a preprocessed signal

    :return: preprocessed data
    """
    
    if ecg_frequency == requested_frequency:
        return ecg_record
    ecg_record = signal.resample(ecg_record, int(ecg_record.shape[1] * requested_frequency / ecg_frequency), axis=1)
    return ecg_record


def DWT_BW(
    s: np.ndarray, 
    wavelet: str = 'db4',
)-> np.ndarray:
    """
    Remove baseline wander using wavelets (see article https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.308.6789&rep=rep1&type=pdf)
    :param s: one lead signal
    :param wavelet: wavelet name

    :return: preprocessed data
    """
    
    w = pywt.Wavelet(wavelet)
    maxlev = pywt.dwt_max_level(len(s), w.dec_len)
    
    diffs = []
    for i in range(1, maxlev + 1):
        coeffs = pywt.wavedec(s, w, level=i, mode='periodic')
        diff = np.sum(np.square(coeffs[0])) - np.sum([np.sum(np.square(coeffs[j])) for j in range(1, i)])
        diffs.append(diff)
    diffs = np.array(diffs)
    if np.max(diffs) > 6500:
        ixs = np.where(diffs > 6500)[0]
        ix = ixs[np.argmin(diffs[ixs])]
        coeffs = pywt.wavedec(s, w, level=ix, mode='periodic')
        coeffs[0] = np.array([0]* len(coeffs[0]))
    else:
        ix = np.argmin(diffs[-3:])
        ix += len(diffs) - 3
        coeffs = pywt.wavedec(s, w, level=ix, mode='periodic')
        coeffs[0] = np.array([0]* len(coeffs[0]))
    return pywt.waverec(coeffs, wavelet, mode='periodic')


def cut_ecg(
    data: np.ndarray, 
    cut_range: list, 
    frequency: int,
) -> np.ndarray:
    """
    Cut signal edges 
    :param data: signal
    :param cut_range: cutting parameters
    :param frequency: sampling frequency of a signal

    :return: preprocessed data
    """
    
    cut_data = []
    start = int(cut_range[0]*frequency)
    for rec in data:
        end = -int(cut_range[1]*frequency) if cut_range[1] != 0 else len(rec)
        cut_data.append(rec[start: end])

    return np.array(cut_data)


import requests
from zipfile import ZipFile
import os

from tqdm import tqdm
import pandas as pd


__all__ = [
    "load_ptb_xl",
]


def load_ptb_xl(download: bool = False,
                path_to_zip: str = "./",
                path_to_unzip: str = "./",
                delete_zip: bool = True,
               ) -> pd.DataFrame:
    '''
    Load PTB-XL dataset
    :param download: whether to download PTB-XL from Physionet
    :param path_to_zip: path where to store PTB-XL .zip file
    :param path_to_unzip: path where to unarchive PTB-XL .zip file
    :param delete_zip: whether to delete PTB-XL .zip file after unarchiving

    :return: dataframe with PTB-XL dataset info
    '''
    
    if download:
        
        url = 'https://physionet.org/static/published-projects/ptb-xl/ptb-xl-a-large-publicly-available-electrocardiography-dataset-1.0.2.zip'
        ptb_xl_zip = os.path.join(path_to_zip, 'ptb_xl.zip')
        response = requests.get(url, stream=True)
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        print('Loading PTB-XL file...')
        with open(ptb_xl_zip, 'wb') as f:
            progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
            for chunk in response.iter_content(chunk_size=512):
                if chunk:
                    progress_bar.update(len(chunk))
                    f.write(chunk)
        progress_bar.close()
        print('Loading completed!')
        f.close()
        
        print('Unzipping PTB-XL file...')
        with ZipFile(ptb_xl_zip, 'r') as zip_ref:
            for member in tqdm(zip_ref.infolist(), desc=''):
                try:
                    zip_ref.extract(member, path_to_unzip)
                except zipfile.error as e:
                    pass
        print('Unzipping completed!')
        
        if delete_zip:
            print(f'Deleting {ptb_xl_zip} file...')
            os.remove(ptb_xl_zip)
            print('Deleting completed!')
            
    ptb_xl_info = pd.read_csv(os.path.join(path_to_unzip, 'ptb-xl-a-large-publicly-available-electrocardiography-dataset-1.0.2', 'ptbxl_database.csv'))
    ptb_xl_info['fpath'] = [os.path.join(path_to_unzip, 'ptb-xl-a-large-publicly-available-electrocardiography-dataset-1.0.2', ptb_xl_info.iloc[i]['filename_hr']) for i in range(len(ptb_xl_info['filename_hr']))]
    
    return ptb_xl_info


import torch
import torch.nn as nn


class DenseLayer(nn.Module):
    ''' Paper: https://arxiv.org/pdf/1608.06993v5.pdf '''
    def __init__(self, input_channels, growth_rate, bottleneck_size, kernel_size):
        super().__init__()
        self.use_bottleneck = bottleneck_size > 0
        self.num_bottleneck_output_filters = growth_rate * bottleneck_size
        if self.use_bottleneck:
            self.bn2 = nn.BatchNorm1d(input_channels)
            self.act2 = nn.ReLU(inplace=True)
            self.conv2 = nn.Conv1d(
                input_channels,
                self.num_bottleneck_output_filters,
                kernel_size=1,
                stride=1)
        self.bn1 = nn.BatchNorm1d(self.num_bottleneck_output_filters)
        self.act1 = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv1d(
            self.num_bottleneck_output_filters,
            growth_rate,
            kernel_size=kernel_size,
            stride=1,
            dilation=1,
            padding=kernel_size // 2)

    def forward(self, x):
        if self.use_bottleneck:
            x = self.bn2(x)
            x = self.act2(x)
            x = self.conv2(x)
        x = self.bn1(x)
        x = self.act1(x)
        x = self.conv1(x)
        return x


class DenseBlock(nn.ModuleDict):
    def __init__(self, num_layers, input_channels, growth_rate, kernel_size, bottleneck_size):
        super().__init__()
        self.num_layers = num_layers
        for i in range(self.num_layers):
            self.add_module(f'denselayer{i}',
                            DenseLayer(input_channels + i * growth_rate,
                                       growth_rate,
                                       bottleneck_size,
                                       kernel_size))

    def forward(self, x):
        layer_outputs = [x]
        for _, layer in self.items():
            x = layer(x)
            layer_outputs.append(x)
            x = torch.cat(layer_outputs, dim=1)
        return x


class TransitionBlock(nn.Module):
    def __init__(self, input_channels, out_channels):
        super().__init__()
        self.bn = nn.BatchNorm1d(input_channels)
        self.act = nn.ReLU(inplace=True)
        self.conv = nn.Conv1d(input_channels, out_channels, kernel_size=1, stride=1, dilation=1)
        self.pool = nn.AvgPool1d(kernel_size=2, stride=2)

    def forward(self, x):
        x = self.bn(x)
        x = self.act(x)
        x = self.conv(x)
        x = self.pool(x)
        return x


class DenseNet1d(nn.Module):
    def __init__(
            self,
            growth_rate: int = 32,
            block_config: tuple = (6, 12, 24, 16),
            num_init_features: int = 64,
            bottleneck_size: int = 4,
            kernel_size: int = 3,
            input_channels: int = 3,
            num_classes: int = 1,
            reinit: bool = True,
    ):
        super().__init__()

        self.features = nn.Sequential(
            nn.Conv1d(
                input_channels, num_init_features,
                kernel_size=7, stride=2, padding=3, dilation=1),
            nn.BatchNorm1d(num_init_features),
            nn.ReLU(inplace=True),
            nn.MaxPool1d(kernel_size=3, stride=2, padding=1),
        )

        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = DenseBlock(
                num_layers=num_layers,
                input_channels=num_features,
                growth_rate=growth_rate,
                kernel_size=kernel_size,
                bottleneck_size=bottleneck_size,
            )
            self.features.add_module(f'denseblock{i}', block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = TransitionBlock(
                    input_channels=num_features,
                    out_channels=num_features // 2)
                self.features.add_module(f'transition{i}', trans)
                num_features = num_features // 2

        self.final_bn = nn.BatchNorm1d(num_features)
        self.final_act = nn.ReLU(inplace=True)
        self.final_pool = nn.AdaptiveAvgPool1d(1)
        self.classifier = nn.Linear(num_features, num_classes)

        # init
        if reinit:
            for m in self.modules():
                if isinstance(m, nn.Conv1d):
                    nn.init.kaiming_normal_(m.weight)
                elif isinstance(m, nn.BatchNorm1d):
                    nn.init.constant_(m.weight, 1)
                    nn.init.constant_(m.bias, 0)
                elif isinstance(m, nn.Linear):
                    nn.init.constant_(m.bias, 0)

    def forward_features(self, x):
        out = self.features(x)
        out = self.final_bn(out)
        out = self.final_act(out)
        out = self.final_pool(out)
        return out

    def forward(self, x):
        features = self.forward_features(x)
        features = features.squeeze(-1)
        out = self.classifier(features)
        return out

    def reset_classifier(self):
        self.classifier = nn.Identity()

    def get_classifier(self):
        return self.classifier


def densenet121_1d(**kwargs):
    kwargs["block_config"] = (6, 12, 24, 16)
    return DenseNet1d(**kwargs)


def densenet201_1d(**kwargs):
    kwargs["block_config"] = (6, 12, 48, 32)
    return DenseNet1d(**kwargs)



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup_args = dict(
    name='ecglib',
    version='1.0.1',
    description='ECG library with pretrained models and tools for ECG analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache License 2.0',
    packages=find_packages(),
    include_package_data=True,
    author='Aram Avetisyan',
    author_email='a.a.avetisyan@gmail.com',
    install_requires=['pandas',
        'ecg_plot',
        'tqdm',
        'numpy',
        'torch',
        'scipy',
        'ipywidgets',
        'pyyaml',
        'PyWavelets',
        'wfdb',
        'hydra-core',
        'omegaconf',
    ],
    keywords=['ecg analysis', 'pytorch', 'pretrained models', 'ecg preprocessing', 'ecg datasets'],
    url = 'https://github.com/ispras/EcgLib',
    python_requires='>=3.6',
)


if __name__ == '__main__':
    setup(**setup_args)



import imp
from dataclasses import dataclass
from typing import Optional, Type

import torch.nn as nn


@dataclass(repr=True, eq=True)
class DenseNetConfig:
    """
    Default parameters correspond DenseNet121_1d model
    """

    growth_rate: int = 32
    block_config: tuple = (6, 12, 24, 16)
    num_init_features: int = 64
    bottleneck_size: int = 4
    kernel_size: int = 3
    input_channels: int = 12
    num_classes: int = 1
    reinit: bool = True